import {
    system,
    world
}
from "@minecraft/server";

import {
    ModalFormData
}
from "@minecraft/server-ui";

const convertRot = (r) => {
    return {
        x: Math.cos(r),
        y: Math.sin(r)
    };
};

const radDegCoeff = Math.PI / 180;

const scaleV2 = (v, s) => {
    return {
        x: v.x * s,
        y: v.y * s
    };
};

const scaleV3 = (v, s) => {
    return {
        x: v.x * s,
        y: v.y * s,
        z: v.z * s
    };
};

const dotV3 = (v0, v1) => {
    return v0.x * v1.x + v0.y * v1.y + v0.z * v1.z;
};

const magV3 = (v) => {
    return Math.hypot(v.x, v.y, v.z);
};

class Drone {
    d;
    p;
    t;
    r;
    l;
    s;
    v;
    o;
    constructor(pltc) {
        this.p = pltc.p;
        if ('d' in pltc) {
            this.d = pltc.d;
            this.loadSave();
        } else {
            this.d = this.p.dimension.spawnEntity(`drond:drone<drond:spawn_${pltc.t}>`, pltc.l);
            this.d.setDynamicProperty("opId", this.p.id);
            this.p.setDynamicProperty("droneId", this.d.id);
            this.t = pltc.t;
            this.l = pltc.l;
            this.r = pltc.p.getRotation();
        }
        this.o = {
            x: 0,
            y: 0,
            z: 0
        };
        this.droneData();
    }

    droneData() {
        switch (this.t) {
        case "scout": {
                this.s = 0.19;
                this.o.y = -0.21;
                break;
            }
        case "suicide": {
                this.s = 0.13;
                this.o.y = 0.5;
                break;
            }
        case "dropper": {
                this.s = 0.15;
                this.o.y = -0.3;
                break;
            }
        case "carrier": {
                this.s = 0.15;
                this.o.y = 0.5;
                break;
            }
        }
        this.updateDrone = (() => {
            switch (this.t) {
            case "carrier": {
                    return (() => {
                        let [u, k] = [this.checkButton("Sneak", "Pressed"), this.checkButton("Jump", "Pressed")];
                        let c = this.d.getComponent("minecraft:rideable");
                        if (c?.getRiders()) {
                            if (u && k) {
                                c.ejectRiders();
                                this.s = 0.15;
                            } else if (u) {}
                        } else {
                            if (u) {
                                let e = this.d.dimension.getEntitiesFromRay(
								this.l,
								{x: 0, y: -1, z: 0},
								{
                                    ignoreBlockCollision: true,
                                    includePassableBlocks: false,
                                    maxDistance: 3,
                                    includeLiquidBlocks: false
                                }).sort((a,b) => {a.distance - b.distance})[0].entity;
                                c.addRider(e);
                                this.s = 0.09;
                            }
                        }
                        let n = u && !k ? -1 : !u && k ? 1 : 0;
                        this.d.applyImpulse({
                            x: 0,
                            y: Math.abs(this.v.y) * n,
                            z: 0
                        });
                        this.o = scaleV3({
                            x: this.v.x,
                            y: 0,
                            z: this.v.z
                        }, -15);
                    });
                    break;
                }
            case "dropper": {
                    return (() => {
                        let u = this.buttonState("Pressed");
                        if (u > 0) {
                            this.d.applyImpulse({
                                x: 0,
                                y: -this.v.y,
                                z: 0
                            })
                        }
                        if (this.pb || (u < 0)) {
                            this.pb = true;
                            if (this.checkButton("Sneak", "Released")) {
                                this.d.applyKnockback(0, 0, 0, 0.15);
                                this.d.triggerEvent("drond:drop_nade");
                                this.pb = false;
                            }
                        }
                    });
                    break;
                }
            case "suicide": {
                    return (() => {
                        let u = this.buttonState("Pressed");
                        this.d.applyImpulse({
                            x: 0,
                            y: Math.abs(this.v.y) * u,
                            z: 0
                        });
                        let b = this.d.getBlockFromViewDirection({
                            maxDistance: 1,
                            includePassableBlocks: false
                        });
                        if (b) {
                            this.d.triggerEvent("drond:explode");
                        }
                    });
                    break;
                }
            case "scout":
            default: {
                    return (() => {
                        let u = this.buttonState("Pressed");
                        this.d.applyImpulse({
                            x: 0,
                            y: Math.abs(this.v.y) * u,
                            z: 0
                        });
                    });
                    break;
                }
            }
        })();
    }

    update() {
        let m = this.p.inputInfo.getMovementVector();
        this.r = this.p.getRotation();
        this.v = this.p.getViewDirection();
        this.v = scaleV3(this.v, this.s);
        let l = {
            x: this.v.z,
            y: 0,
            z: -this.v.x
        };
        if (m.y > 0) {
            this.d.applyImpulse(this.v)
        } else if (m.y < 0) {
            this.d.applyImpulse(scaleV3(this.v, -1))
        }
        if (m.x > 0) {
            this.d.applyImpulse(l)
        } else if (m.x < 0) {
            this.d.applyImpulse(scaleV3(l, -1))
        }
        this.updateDrone();
        this.display();
    }

    checkButton(b, s) {
        return this.p.inputInfo.getButtonState(b) === s;
    }

    buttonState(s) {
        return this.checkButton("Jump", s) ? 1 : this.checkButton("Sneak", s) ? -1 : 0
    }

    display() {
        this.l = this.d.location;
        this.p.onScreenDisplay.setActionBar(`Location: ${this.l.x.toFixed()} ${this.l.y.toFixed()} ${this.l.z.toFixed()}`);
        this.d.setRotation(this.r);
        this.p.camera.setCamera("minecraft:free", {
            location: {
                x: this.l.x + this.o.x,
                y: this.l.y + this.o.y,
                z: this.l.z + this.o.z
            },
            rotation: this.r,
            easeOptions: {
                easeTime: 1.7,
                easeType: "OutExpo"
            }
        });
    }

    unset() {
        this.p.camera.clear();
        this.p.inputPermissions.movementEnabled = true;
    }

    unload() {
        loaded.delete(this.p.id);
        unloaded.set(this.p.id, this);
        system.runTimeout(() => {
            this.unset();
        }, 3);
    }

    set() {
        this.p.inputPermissions.movementEnabled = false;
    }

    load() {
        unloaded.delete(this.p.id);
        loaded.set(this.p.id, this);
        this.set();
    }

    save() {
        this.p.setDynamicProperty("location", this.l);
        this.p.setDynamicProperty("droneType", this.t);
        this.p.setDynamicProperty("droneId", this.d.id);
    }

    loadSave() {
        this.l = this.p.getDynamicProperty("location");
        this.t = this.p.getDynamicProperty("droneType");
        this.d.setDynamicProperty("opId", this.p.id);
    }

    kill() {
        this.d.kill();
    }

    tick() {
        this.d.triggerEvent("drond:tick");
    }

    untick() {
        this.d.triggerEvent("drond:untick");
    }
}

const loaded = new Map();
const unloaded = new Map();

const useDrone = {
    onUse(e) {
        let p = e.source;
        if (p.inputInfo.getButtonState("Sneak") === "Pressed") {
            settings(p);
        } else {
            let [ld, ud] = [loaded.get(p.id), unloaded.get(p.id)];
            if (ld) {
                ld.unload();
            } else if (ud) {
                ud.load();
            } else {
                let o = p.getDynamicProperty("offsetXYZ");
                let d = new Drone({
                    p: p,
                    l: {
                        x: p.location.x + o.x,
                        y: p.location.y + o.y,
                        z: p.location.z + o.z
                    },
                    t: p.getDynamicProperty("droneType")
                });
                if (p.getDynamicProperty("switch")) {
                    d.load()
                } else {
                    d.unload()
                }
                d.tick();
            }
        }
    }
};

const settings = (p) => {
    const types = new Map();
    types.set(0, "scout");
    types.set(1, "suicide");
    types.set(2, "dropper");
    types.set(3, "carrier");
    let f = settings_form(p);
    f.show(p).then(r => {
        if (r.canceled)
            return;
        let [t, s, y, z, x, k] = r.formValues;
        p.setDynamicProperty("droneType", types.get(t));
        p.setDynamicProperty("switch", s);
        p.setDynamicProperty("offsetXYZ", {
            x: x,
            y: y,
            z: z
        });
        if (k) {
            let [ld, ud] = [loaded.get(p.id), unloaded.get(p.id)];
            if (ld) {
                ld.kill()
            } else if (ud) {
                ud.kill()
            }
        }
    }).catch((e) => {
        console.error(e);
    });
};

const settings_form = (p) => {
    const types = new Map();
    types.set("scout", 0);
    types.set("suicide", 1);
    types.set("dropper", 2);
    types.set("carrier", 3);
    let f = new ModalFormData();
    f.title("Drone Settings");
    f.dropdown("Drone Type", Array.from(types.keys()), types.get(p.getDynamicProperty("droneType")));
    f.toggle("Switch To Drone Camera On Spawn", p.getDynamicProperty("switch"));
    let o = p.getDynamicProperty("offsetXYZ");
    f.slider("Y Drone Spawn Offset", -1, 5, 1, o.y);
    f.slider("Z Drone Spawn Offset", -1, 5, 1, o.z);
    f.slider("X Drone Spawn Offset", -1, 5, 1, o.x);
    f.toggle("Kill Drone", false);
    return f;
};

world.beforeEvents.worldInitialize.subscribe(e => {
    e.itemComponentRegistry.registerCustomComponent("drond:use_drone", useDrone);
});

world.beforeEvents.playerLeave.subscribe(({
        player
    }) => {
    let [ld, ud] = [loaded.get(player.id), unloaded.get(player.id)];
    if (ld) {
        ld.save();
        ld.unload();
        system.runTimeout(() => {
            ld.untick();
        }, 2);
    } else if (ud) {
        ud.save();
        system.runTimeout(() => {
            ud.untick();
        }, 2);
    }
});

world.afterEvents.playerSpawn.subscribe(({
        initialSpawn,
        player
    }) => {
    let d = unloaded.get(player.id);
    if (d) {
        d.tick();
    } else {
        let did = player.getDynamicProperty("droneId");
        if (did) {
            let d = world.getEntity(did);
            if (d) {
                let dn = new Drone({
                    p: player,
                    d: d
                });
                dn.unload();
                dn.tick();
            }
        }
    }
});

world.beforeEvents.entityRemove.subscribe(({
        removedEntity
    }) => {
    if (removedEntity.typeId === "drond:drone") {
        let pid = removedEntity.getDynamicProperty("opId");
        let [ld, ud] = [loaded.get(pid), unloaded.get(pid)];
        if (loaded.delete(pid)) {
            system.runTimeout(() => {
                ld.unset()
            }, 3);
        } else {
            unloaded.delete(pid)
        }
    }
});

system.runInterval(() => {
    for (let [_, d] of loaded) {
        d.update();
    }
}, 1);
